#include <stdlib.h>
#include <stdio.h>
#include <dos.h>

#include "trdir.h"
#include "trmsg.h"
#include "trapp.h"

#define DIRECTSTATION   1
#define Swap(num)       ((num) << 8) + ((num) >> 8)
#define MAXQUEUES       2
#define FREEQ           0
#define MSGQ            1

#define BUFFERS         128

/* MAC frame vector structure */
typedef struct {
	WORD VL;		//vector length
	BYTE VIC;		//vector source and destination class
	BYTE VI;		//vector identifier
	BYTE Data[1024];	//vector value
	} MajVector;

/* MAC frame subvector structure */
typedef struct {
	BYTE SVL;		//subvector length
	BYTE SVI;		//subvector identifier
	BYTE Data[256];		//subvector data
	} SubVector;

/* Internal queue structure */
typedef struct {
Msgs            *pHead;         // head of queue 
Msgs            *pTail;         // tail of queue
int             count;          // queue count
} Queue_t;


/* global variables */
Queue_t queue[MAXQUEUES];		//message queues
CCB workccb, bufferccb, recvccb;	//token ring command control blocks
BYTE local_addr[6];			//token ring node address
BYTE adapter;				//token ring adapter number
WORD last_status;			//last ring status
Msgs app_msg;				//message to pass back to TRMGR
int initialized;			//flag to record if TRAGN is in use
Msgs msgpool[BUFFERS];			//message buffer pool

/* function prototypes */
static BYTE AdapterInit(void);
static BYTE Activate(void);
static void StatAppFn(unsigned status);
static void RecvAppFn(void * ptr);
static int ValidInfo(MajVector *ptr, Msgs *msg);
static Msgs *IntServiceFn(unsigned code);
static void Put(short idQueue, Msgs * pItem);
static Msgs * Get(short idQueue);
static short Qsize(short idQueue);
static void FreeMsgs(void);


/* Queue an item in the specified queue */
static void Put(short idQueue, Msgs * pItem)
{
   Queue_t *pQueue;
   
   if (idQueue < MAXQUEUES)
      {
      pQueue = &queue[idQueue];
      if (pQueue->pHead == NULL)
	pQueue->pHead = pItem;
      else
	pQueue->pTail->next = pItem;
	
      pQueue->pTail = pItem;
      pItem->next = NULL;
      pQueue->count++;
      }  
}

/* Dequeue an item from the specified queue */
static Msgs * Get(short idQueue)
{
   Queue_t *pQueue;
   Msgs *pItem = NULL;
   
   if (idQueue < MAXQUEUES)
      {
      pQueue = &queue[idQueue];
      pItem = pQueue->pHead;
      if (pQueue->pHead != NULL)
	{
	pQueue->pHead = pQueue->pHead->next;
	if (pQueue->pHead == NULL)
	  pQueue->pTail = NULL;
	pQueue->count--;
	}
      }
      
   /* if no more free buffers, get one from the msg queue ie, discard the
      oldest message */
   if ((pItem == NULL) && (idQueue == FREEQ))
	pItem = Get(MSGQ);
	
   return (pItem);  
}


/* Returns size of queue */
static short Qsize(short idQueue)
{
   if (idQueue < MAXQUEUES)
      return queue[idQueue].count;
   else
      return 0;
      
}


/* free all queued messages */
static void FreeMsgs(void)
{
	Msgs *queue;

	while ((queue = Get(MSGQ)) != NULL)
	   {
	   buffer_free(&bufferccb, adapter, 0, (BUFFER_PT *) queue->buffer_one);
	   Put(FREEQ, queue);
	   }
}


/* ring status appendage function */
static void StatAppFn(unsigned status)
{
	Msgs * msg;
	
	if (!initialized) return;

	/* only create and queue a message if TRAGN is in use */
	status &= ~(STS_HARDERROR | STS_SOFTERROR );
	if (last_status != status)
	   {
	   last_status = status;
	   /* create message */
	   msg = Get(FREEQ);
	   msg->msg = MSG_DATA_INDICATION;
	   msg->ringstatus = status;
	   Put(MSGQ, msg);
	   }
}


/* data received appendage function */
static void RecvAppFn(void *ptr)
{

	RECV_BUFFER_ONE *BufOne;
	RECV_BUFFER_TWO *BufTwo;
	MajVector *MVptr;
	Msgs *msg;
	CCB *ccb;

	BufOne = (RECV_BUFFER_ONE *) ptr;

	/* only process MAC frames */
	if (BufOne != (RECV_BUFFER_ONE *) ZEROADDRESS)
	  {
	  if ( BufOne->msg_type == 0x02)
	    {
	    /* Since we used the BREAK option in receive, the MAC vector
		is always contained in the second buffer */

	    BufTwo = (RECV_BUFFER_TWO *) BufOne->next_buf_ptr;
	    if (BufTwo != (RECV_BUFFER_TWO *) ZEROADDRESS)
	      {
	      MVptr = (MajVector *) &BufTwo->rcvd_data;
  
	      msg = Get(FREEQ);
	      msg->buffer_one = (ADDRESS) BufOne;

	      /* check if the MAC frame is the one we are looking for and
		 create and queue it as a message */
	      if (ValidInfo(MVptr, msg))
		 {
		 memcpy(msg->adapter, BufOne->lan_hdr.srcaddr, 6);
		 Put(MSGQ, msg);
		 }
	      else
		 {
		 msg->msg = MSG_HOUSEKEEPING;
		 Put(MSGQ, msg);
		 }
  
	    }
	  }
  
       }


}


/* check if the MAC frame vector and subvectors are the one we are 
   interested */
static int ValidInfo(MajVector *ptr, Msgs *msg)
{
    int len;
    int i = 0;
    SubVector *subvec;

    /* associate vector with message */
    switch (ptr->VI) {
       case 0x05:
       case 0x06:
       case 0x26:
	  msg->msg = MSG_DATA_CONFIG;
	  break;
       case 0x29:
	  msg->msg = MSG_DATA_ERROR;
	  break;
       case 0x02:
	  msg->msg = MSG_DATA_BEACON;
	  break;
       default:
	  return 0;
    }

    /* go through all the subvectors contained in the vector. Note that
	the length of the vector is in big-endian format. */
    len = Swap(ptr->VL);
    while (i < ( len - 4 ))
	{
	subvec = (SubVector *) &ptr->Data[i];
	switch ( subvec->SVI ) {
	    case 0x01:
		msg->beacontype = subvec->Data[1];
		break;
	    case 0x02:
		memcpy((void *) msg->naun, (void *) subvec->Data, 6);
		break;
	    case 0x2d:
		memcpy((void *) msg->isoerr, 
		  (void *) subvec->Data, 5);
		break;
	    case 0x2e:
		memcpy((void *) msg->noniso, 
		  (void *) subvec->Data, 5);
		break;
	    default:
		break;

		}
	i += subvec->SVL;
	}

    return (1);
}


/* TRAGN's service interrupt function */
static Msgs *IntServiceFn(unsigned code)
{

	Msgs *queue;
	ADAPTER_STATUS_PT *status;
	Msgs *msg;

	msg = &app_msg;
	msg->msg = code;
	msg->retcode = ERR_OK;

	/* service different requests from TRMGR */
	switch (code)
	  {
	  case MSG_REQUEST_MSG:
		/* free all direct buffers back to the direct buffer pool
		   for housekeeping messages */
		while (((queue = Get(MSGQ)) != NULL) &&
		       (queue->msg == MSG_HOUSEKEEPING))
		   {
		   buffer_free(&bufferccb, adapter, 0, (BUFFER_PT *) queue->buffer_one);
		   Put(FREEQ, queue);
		   }

		/* copy the message and release the buffers */
		if (queue != NULL)
		  {
		  *msg = *queue;
		  buffer_free(&bufferccb, adapter, 0, (BUFFER_PT *) queue->buffer_one);
		  Put(FREEQ, queue);
		  }
		else
		  msg->retcode = ERR_NO_DATA;
		break;

	  case MSG_REQUEST_RESET:
		FreeMsgs();
		int_on();
		if ((msg->dir_retcode = AdapterInit()) != 0)
		   msg->retcode = ERR_FAILURE;
		else
		   memcpy((void *) msg->adapter, (void *) local_addr, 6);
		break;


	  case MSG_REQUEST_REMOVE:
		int_on();
		if ((msg->dir_retcode = RemoveStn(adapter, DIRECTSTATION, 
		     msg->adapter)) != 0)
		   msg->retcode = ERR_FAILURE;
		break;

	  case MSG_REQUEST_INACTIVATE:
		int_on();
		if ((msg->dir_retcode = receive_cancel(&recvccb, adapter, 
		     DIRECTSTATION)) != 0)
		   msg->retcode = ERR_FAILURE;
		else
		   memcpy((void *) msg->adapter, (void *) local_addr, 6);
		break;
		initialized = 0;
		FreeMsgs();

	  case MSG_REQUEST_ACTIVATE:
		int_on();
		if ((msg->dir_retcode = Activate()) != 0)
		   msg->retcode = ERR_FAILURE;
		else
		   memcpy((void *) msg->adapter, (void *) local_addr, 6);
		break;

	  default:
		msg->retcode = ERR_INVALID_MSG;
		break;
	  }
	
	msg->more = Qsize(MSGQ) & 0xff;

	return msg;
}


/* initialize the adapter */
static BYTE AdapterInit(void)
{
	/* see if adapter has been initialized */
	dir_interrupt( &workccb, adapter );
	BusyWait(workccb.retcode);

	if ( workccb.retcode && (workccb.retcode != 0x09 ))
		return workccb.retcode;

	/* initialize the adapter */
	dir_initialize( &workccb, adapter );
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return workccb.retcode;

	/* open the adapter */
	dir_openadapter( &workccb, adapter, StaApp);
	BusyWait(workccb.retcode);

	switch (workccb.retcode){
	   case 0:
	   case 2:
	   case 3:
	   case 0x10:
		break;
	   default:
		return workccb.retcode;
	   }


	/* set functional addresses */
	dir_set_func_address(&workccb, adapter, 0x18000000L);
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return workccb.retcode;


	/* issue receive to receive MAC frames */
	receive_cancel(&recvccb, adapter, DIRECTSTATION);
	BusyWait(workccb.retcode);

	receive(&recvccb, adapter, DIRECTSTATION, RecApp);
	if (recvccb.retcode != 0xff)
		return recvccb.retcode;

	/* get our adapter's address */
	dir_status( &workccb, adapter );
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return workccb.retcode;

	memcpy((void *)  local_addr, 
		(void *) ((STATUS_PT *)workccb.parm_tab)->node_addr, 6);
	initialized = 1;

	return 0;
}


/* See if the adapter has already been set up in a previous session. 
   If so skip the initialization part */
static BYTE Activate(void)
{
	/* see if adapter has been initialized */
	dir_interrupt( &workccb, adapter );
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return AdapterInit();

	/* open adapter: note that the ring status appendage address is 
	   that provided in TRAPP, not StatAppFn */
	dir_openadapter( &workccb, adapter, StaApp);
	BusyWait(workccb.retcode);

	if  (workccb.retcode == 0x03)
		{
		dir_modify_open_parms( &workccb, adapter, StaApp);
		BusyWait(workccb.retcode);

		if (workccb.retcode && (workccb.retcode != 0x02)) 
			return workccb.retcode;
		}
	else if (workccb.retcode && (workccb.retcode != 0x10))
		return workccb.retcode;


	/* set functional addresses */
	dir_set_func_address(&workccb, adapter, 0x18000000L);
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return workccb.retcode;


	/* issue receive: note that the data received appendage address is 
	   that provided in TRAPP, not RecvAppFn */
	receive(&recvccb, adapter, DIRECTSTATION, RecApp);
	if (recvccb.retcode != 0xff)
		return recvccb.retcode;

	/* get our adapter's address */
	dir_status( &workccb, adapter );
	BusyWait(workccb.retcode);

	if (workccb.retcode)
		return workccb.retcode;

	memcpy((void *)  local_addr, 
		(void *) ((STATUS_PT *)workccb.parm_tab)->node_addr, 6);
	initialized = 1;

	return 0;
}


/* do initialization and stay resident */
void main(void)
{

	Msgs * ptr;
	int i;
	union REGS inregs, outregs;
	struct SREGS segregs;
	unsigned *seg;
	unsigned program_end;
	
	/* set up internal queue for messages */
	adapter = 0;
	ptr = (Msgs *) &msgpool;
	for (i = 0; i < BUFFERS; i++)
		Put(FREEQ, ptr++);      

	/* check if the token ring API support has been installed */
	if (!Installed(adapter)) {
		printf( "Error: Token Ring adapter Support Interface not loaded!\n" );
		return;
	}


	/* install interrupt service routine to serve the TRMGR's requests.
	   This call registers the 3 functions with the TRAPP appendage 
	   support */
	appinit(IntServiceFn, RecvAppFn, StatAppFn);

	inregs.x.dx  = FP_OFF(SrvApp);
	segregs.ds = FP_SEG(SrvApp);
	inregs.x.ax = 0x2500 + SERVINT;
	intdosx(&inregs, &outregs, &segregs);

	/* release environment */
	seg =  MK_FP(_psp, 0x2c);
	dos_free(*seg);

	printf("Installing Token Ring Agent V1.0\n");

	/* determine size of program */
	program_end = dos_alloc(256);
	dos_free(program_end);
	inregs.x.dx  = program_end  - _psp + 1;
	inregs.x.ax = 0x3100;
	segread(&segregs);

	/* inform dos of our new size */
	dos_setblock(inregs.x.dx, _psp);
	
	/* stay resident */
	intdos(&inregs, &outregs);

	printf("Failed to install agent, error %d\n",outregs.h.al);


}


